/**
 * @file usysident.c
 * @brief usysident/usysattn/usysfault 
 */
/**
 * @mainpage usysident/usysattn documentation
 * @section Copyright
 * Copyright (c) 2004 International Business Machines
 * Common Public License Version 1.0 (see COPYRIGHT)
 *
 * @section Overview
 * Manipulate LEDs on IBM ppc64 platforms.
 *
 * @section usysident
 * The usysident utility is used to view and manipulate the indicators
 * (LEDs) that identify certain devices on PowerPC-64 machines.  These
 * identifiers are specified by location code; location codes can be retrieved
 * by running lscfg.
 *
 * When run without arguments, usysident will print a list of all of the
 * identification indicators on the system along with their current status (on
 * or off).  The <tt>-l</tt> or <tt>-d</tt> options can be used to specify a
 * particular indicator, by location code or logical device name respectively.
 * If <tt>-l</tt> or <tt>-d</tt> is the only argument, the status of that 
 * indicator will be printed.  If the <tt>-s</tt> argument is specified in 
 * addition, the indicator may be turned on or off.
 *
 * @section usysattn
 * The usysattn utility is used to view and manipulate the attention
 * indicators (LEDs) on PowerPC-64 machines.  These indicators are turned on
 * automatically when a serviceable event is received.  These identifiers are
 * specified by location code.
 * 
 * When run without arguments, usysattn will print a list of all of the
 * attention indicators on the system along with their current status (on or
 * off).  The <tt>-l</tt> option can be used to specify a particular indicator.
 * If <tt>-l</tt> is the only argument, the status of that indicator will be
 * printed.  If <tt>-s normal</tt> is specified along with the <tt>-l</tt>
 * argument, the indicator will be turned off.
 * 
 * <b>NOTE:</b> Ensure that there are no outstanding service actions to be 
 * taken on this system before turning off the system attention indicators.
 *
 * @author Michael Strosaker <strosake@us.ibm.com>
 */

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <librtas.h>
#include <ctype.h>

#include "librtas_error.h"

#define BUF_SIZE	4096
#define IDENT_INDICATOR	9007	/**< usysident indicator */
#define ATTN_INDICATOR	9006	/**< usysattn indicator */

#define INDICATOR_TYPE(x)	(((x) == IDENT_INDICATOR) ? "identification" \
							  : "attention")

/**
 * @struct loc_code
 * @brief location code definition
 *
 * The order of the fields in loc_code is important! 
 */
struct loc_code {
	uint32_t length;	/**< length includes null terminator */
	char code[120];
	uint32_t index;
	struct loc_code *next;
};

/**
 * truncate_loc_code
 * @brief truncate the last few characters of a location code
 *
 * Truncates the last few characters off of a location code; if an
 * indicator doesn't exist at the orignal location, perhaps one exists
 * at a location closer to the CEC.
 *
 * @param loccode location code to truncate
 * @return 1 - successful truncation
 * @return 0 - could not be truncated further
 */
int
truncate_loc_code (char *loccode) {
	int i;

	for (i = strlen(loccode) - 1; i >= 0; i--) {
		if (loccode[i] == '-') {
			loccode[i] = '\0';
			return 1;	/* successfully truncated */
		}
	}
	return 0;	/* could not be further truncated */
}

/**
 * print_usage
 * @brief print the usage statement
 *
 * @param cmd command we are running
 */
void
print_usage (char *cmd) {
	if (strstr(cmd, "usysident"))
		printf("Usage: %s"
			" [-s {normal|identify}] [-l location_code]\n",
			cmd);
	else
		printf("Usage: %s"
			"  [-s normal] [-l location_code]\n", cmd);
}

/**
 * get_sensor
 * @brief retrieve a sensor value from rtas
 * 
 * Call the rtas_get_sensor or rtas_get_dynamic_sensor librtas calls,
 * depending on whether the index indicates that the sensor is dynamic.
 *
 * @param indicator sensor to get the value of
 * @param loc location code of the sensor
 * @param state return location for the sensor state
 * @return rtas call return code
 */
int
get_sensor (int indicator, struct loc_code *loc, int *state) {
	int rc;
	char err_buf[40];

	if (loc->index == 0xFFFFFFFF)
		rc = rtas_get_dynamic_sensor (indicator,
				(void *)loc, state);
	else
		rc = rtas_get_sensor (indicator, loc->index, state);

	if (rc == -1) {
		fprintf(stderr,
			"Hardware error retrieving the indicator at %s\n",
			loc->code);
		return -1;
	}
	else if (rc == -3) {
		fprintf(stderr,
			"The indicator at %s is not implemented.\n",
			loc->code);
		return -2;
	}
	else if (rc != 0) {
		librtas_error(rc, err_buf, 40);

		fprintf(stderr, "Could not get %ssensor %s indicators,\n%s.\n",
			(loc->index == 0xFFFFFFFF) ? "dynamic " : "",
			INDICATOR_TYPE(indicator), err_buf);

		return -3;
	}

	return rc;
}

/**
 * set_indicator
 * @brief set an rtas indicator
 * 
 * Call the rtas_set_indicator or rtas_set_dynamic_indicator librtas calls,
 * depending on whether the index indicates that the indicator is dynamic.
 *
 * @param indicator rtas indicator to set
 * @param loc location code of rtas indicator
 * @param new_value value to update indicator to
 * @return rtas call return code
 */
int
set_indicator (int indicator, struct loc_code *loc, int new_value) {
	int rc;
	char err_buf[40];

	if (loc->index == 0xFFFFFFFF)
		rc = rtas_set_dynamic_indicator (indicator,
			new_value, (void *)loc);
	else
		rc = rtas_set_indicator (indicator, loc->index, new_value);

	if (rc == -1) {
		fprintf(stderr,
			"Hardware error retrieving the indicator at %s\n",
			loc->code);
		return -1;
	}
	else if (rc == -3) {
		fprintf(stderr,
			"The indicator at %s is not implemented.\n",
			loc->code);
		return -2;
	}
	else if ( rc!= 0) {
		librtas_error(rc, err_buf, 40);

		fprintf(stderr, "Could not set %ssensor %s indicator,\n%s.\n",
			(loc->index == 0xFFFFFFFF) ? "dynamic " : "",
			INDICATOR_TYPE(indicator), err_buf);

		return -3;
	}

	return rc;
}

/**
 * delete_list
 * @brief free the memory for list of loc_code structures
 *
 * @param l list to free
 */
void
delete_list(struct loc_code *l) {
	if (!l) return;

	if (l->next) delete_list(l->next);
	free(l);

	return;
}

#ifdef DEBUG
/**
 * dump_raw
 * @brief dump the raw contents of a buffer
 *
 * @param data buffer to dump
 * @param data_len length of the data buffer
 */ 
void 
dump_raw_data(char *data, int data_len) {
	int i, j;
	int offset = 0;
	char *h, *a;
	char *end = data + data_len;

	h = a = data;

	while (h < end) {
		printf("0x%08x  ", offset);
		offset += 16;

		for (i = 0; i < 4; i++) {
			for (j = 0; j < 4; j++) {
				if (h <= end)
					printf("%02x", *h++);
				else
					printf("  ");
			}
			printf(" ");
		}

		printf("|");
		for (i = 0; i < 16; i++) {
			if (a <= end) {
				if ((*a >= ' ') && (*a <= '~'))
					printf("%c", *a);
				else
					printf(".");
				a++;
			} else
				printf(" ");
		}
		printf("|\n");
	}
}
#endif

/**
 * legalize_path
 * @brief Removes any ".." directories from a path
 * 
 * Removes any ".." directories from a path, so that
 * /proc/device-tree/rtas/../../ppc64/rtas/error_log would become
 * /proc/ppc64/rtas/error_log.
 *
 * @param s path to legalize
 */
void
legalize_path(char *s) {
	int p1 = 0, p2 = 0;

	while(s[p1]) {
		if (s[p1] == '.' && s[p1+1] == '.' && s[p1-1] == '/') {
			/* move p2 back one directory */
			p2 -= 2;
			while (s[p2] != '/')
				p2--;
			p1 += 2;
		}
		s[p2++] = s[p1++];
	}
	s[p2] = '\0';

	return;
}

/**
 * parse_work_area
 * @brief parse the working buffer into alist of loc_code structs
 *
 * @param c list to add new data to
 * @param buf worlking area to parse
 * @return pointer to new loc_code list
 */
struct loc_code *
parse_workarea(struct loc_code *c, char *buf) {
	int i;
	struct loc_code *ret = c, *curr = c;
	uint32_t num = *(uint32_t *)buf;

	if (curr)
		while (curr->next)
			curr=curr->next;

	buf += sizeof(uint32_t);
	for (i=0; i<num; i++) {
		if (!curr) {
			curr =
			  (struct loc_code *)malloc(sizeof(struct loc_code));
			ret = curr;
		}
		else {
			curr->next =
			  (struct loc_code *)malloc(sizeof(struct loc_code));
			curr = curr->next;
		}
		curr->index = *(uint32_t *)buf;
		buf += sizeof(uint32_t);
		curr->length = *(uint32_t *)buf;
		buf += sizeof(uint32_t);
		strncpy(curr->code, buf, curr->length);
		buf += curr->length;
		curr->code[curr->length] = '\0';
		curr->length = strlen(curr->code) + 1;
		curr->next = NULL;
	}

	return ret;
}

int
main (int argc, char **argv)
{
	char *dvalue=NULL, *lvalue=NULL, *svalue=NULL, *othervalue=NULL;
	int c, index, next_index, indicator, rc, state, fd;
	char workarea[BUF_SIZE], err_buf[40], loc[120], path[256];
	char sym[128], temp[128];
	struct loc_code *current = NULL, *list = NULL;
	DIR *dir1, *dir2;
	int trunc = 0, truncated = 0;

	while ((c = getopt (argc, argv, "td:l:s:-:")) != -1) {

		switch (c) {
		case 'd':
			dvalue = optarg;
			break;
		case 'l':
			lvalue = optarg;
			break;
		case 's':
			svalue = optarg;
			break;
		case 't':
			trunc = 1;
			break;
		case '-':
			othervalue = optarg;
			break;
		case '?':
			if (isprint (optopt))
				fprintf(stderr,
					"Unrecognized option: -%c\n",
					optopt);
			else
				fprintf(stderr,
					"Unrecognized option character %x\n",
					optopt);
			print_usage (argv[0]);
			return 1;
		default:
			abort ();
		}
	}

	/* Option checking */
	for (index = optind; index < argc; index++) {
		fprintf(stderr,
			"Unrecognized argument %s\n", argv[index]);
		print_usage (argv[0]);
		return 1;
	}

	if (dvalue && !strstr(argv[0], "usysident")) {
		fprintf(stderr,
			"Unrecognized option: -d\n");
		print_usage (argv[0]);
		return 1;
	}

	if (dvalue && lvalue) {
		fprintf(stderr,
			"The -d and -l options cannot be used together.\n");
		print_usage (argv[0]);
		return 1;
	}

	if (strstr(argv[0], "usysident") && svalue
			&& (strcmp(svalue, "identify")
			&& (strcmp(svalue, "normal")))) {
		fprintf(stderr,
			"The -s option must be either "
			"\"identify\" or \"normal\".\n");
		print_usage (argv[0]);
		return 1;
	}
	else if ((strstr(argv[0],"usysattn") || strstr(argv[0],"usysfault"))
		&& svalue &&  strcmp(svalue, "normal")) {
		fprintf(stderr,
			"The -s option must be \"normal\".\n");
		print_usage (argv[0]);
		return 1;
	
	}
	
	if (svalue && !(dvalue || lvalue)) {
		if (strstr(argv[0],"usysident"))
			fprintf(stderr,
				"The -s option requires the -d or -l "
				"option to also be used.\n");
		else
			fprintf(stderr,
				"The -s option required the -l option "
				"to also be used.\n");
		print_usage (argv[0]);
		return 1;
	}

	if (othervalue && strcmp(othervalue, "all-on")
			&& strcmp(othervalue, "all-off")) {
		fprintf(stderr,
			"Unrecognized option: --%s\n", othervalue);
		print_usage (argv[0]);
		return 1;
	}

	if (othervalue && !strcmp(othervalue, "all-on") &&
			!strstr(argv[0], "usysident")) {
		fprintf(stderr,
			"Unrecognized option: --all-on\n");
		print_usage (argv[0]);
		return 1;
	}

	if (othervalue && argc > 2) {
		fprintf(stderr,
			"--%s cannot be used with any other options.\n",
			othervalue);
		print_usage (argv[0]);
		return 1;
	}

	if (strstr(argv[0], "usysident"))
		indicator=IDENT_INDICATOR;
	else if (strstr(argv[0],"usysattn") || strstr(argv[0],"usysfault"))
		indicator= ATTN_INDICATOR;
	else
		return(1);

	index = 1;
	do
	{
		rc = rtas_get_indices (0, indicator, workarea, BUF_SIZE,
				index, &next_index);
		switch(rc) {
		case 1:		/* more data available */
			index = next_index;
			/* fall through */
			
		case 0:		/* success */
			list = parse_workarea(list, workarea);
			break;
			
		case -1:	/* hardware error */
			fprintf(stderr, "Hardware error retrieving "
				"indicator indices\n");
			delete_list(list);
			return -1;
			
		case RTAS_UNKNOWN_OP:
			/* Yes, this is a librtas return code but it should
			 * be treated the same as a -3 return code, both
			 * indicate that functionality is not supported
			 */
			librtas_error(rc, err_buf, 40);
			/* fall through */

		case -3:	/* indicator type not supported. */ 
			fprintf(stderr, "The %s indicators are not supported "
				"on this system", INDICATOR_TYPE(indicator));
			
			if (rc == RTAS_UNKNOWN_OP)
				fprintf(stderr, ",\n%s\n", err_buf);
			else
				fprintf(stderr, ".\n");

			delete_list(list);
		
			return -2;

		case -4:	/* list changed; start over */
			delete_list(list);
			list = NULL;
			index = 1;
			break;

		default:
			librtas_error(rc, err_buf, 40);
			fprintf(stderr, "Could not retrieve data for %s "
				"indicators,\n%s.\n", 
				INDICATOR_TYPE(indicator), err_buf);
			delete_list(list);
			return -3;
		}

	} while (((rc == 1) || (rc == -4)));

	if (!svalue && !lvalue && !dvalue && !othervalue) {
		/* Print all location codes and indicator status */
		current = list;
		while (current) {
			if (!get_sensor(indicator, current, &state))
				printf("%s\t[%s]\n", current->code,
					state?"on":"off");
				fflush(stdout);
			current	= current->next;
		}
	}

	if (dvalue) {
		/* Print/update indicator status for the specified device */

		/* Retrieve the devspec from sysfs */
		/* Look in /sys/class/net and /sys/block */
		if (((dir1 = opendir("/sys")) == NULL) ||
				((dir2 = opendir("/proc")) == NULL)) {
			fprintf(stderr, "The -d option requires a version "
				"2.6 or later kernel with both /sys and "
				"/proc mounted.\n");
			delete_list(list);
			return 1;
		}
		closedir(dir1);
		closedir(dir2);

		workarea[0] = '\0';

		/* Look in /sys/class/net and /sys/block */
		strcpy(path, "/sys/class/net/");
		strcat(path, dvalue);
		strcat(path, "/device/devspec");
		fd = open(path, O_RDONLY);
		if (fd > 0) {
			c = read(fd, workarea, BUF_SIZE);
			close(fd);
			workarea[c] = '\0';
		}
		else {
			strcpy(path, "/sys/block/");
			strcat(path, dvalue);
			fd = open(path, O_RDONLY);
			if (fd > 0) {
				close(fd);
				strcpy(temp, path);
				strcat(temp, "/device");
				if((c = readlink(temp, sym, 128)) > 0) {
					sym[c] = '\0';
					strcat(path, "/");
					strcat(path, sym);
					strcat(path, "/../../devspec");
					legalize_path(path);

					fd = open(path, O_RDONLY);
					if(fd > 0) {
						c = read(fd, workarea,
							BUF_SIZE);
						close(fd);
						workarea[c] = '\0';
					}
				}
			}
		}

		if (workarea[0] == '\0') {
			fprintf(stderr, "Could not determine the location "
				"code for the specified device %s.\n", dvalue);
			delete_list(list);
			return -2;
		}

		/* Retrieve the location code from procfs */
		strcpy(path, "/proc/device-tree");
		strcat(path, workarea);
		strcat(path, "/ibm,loc-code");
		fd = open(path, O_RDONLY);
		if (fd < 0) {
			fprintf(stderr, "Could not determine the location "
				"code for the specified device %s.\n", dvalue);
			delete_list(list);
			return -2;
		}
		c = read(fd, loc, 120);
		close(fd);
		loc[c] = '\0';

		/* Set lvalue and fall through to the next conditional */
		lvalue = loc;
		printf("%s is at location code %s.\n", dvalue, lvalue);
	}

	if (lvalue) {
		/* Print/update indicator status at the specified location */
		if (svalue) {
			if (!strcmp(svalue, "normal"))
				c = 0;
			else if (!strcmp(svalue, "identify"))
				c = 1;
			else {
				fprintf(stderr, "oops! the -s value was "
					"neither \"normal\" nor "
					"\"identify\"!\n");
				delete_list(list);
				return 1;
			}
		}

retry:
		current = list;

		while (current) {
			if (!strcmp(current->code, lvalue)) {
				if (truncated)
					printf("Truncated the specified "
						"location code to %s\n",
						lvalue);

				if (svalue) {
					rc = get_sensor(indicator,
						current, &state);
					if (rc) {
						delete_list(list);
						return rc;
					}

					if (state != c) {
						rc = set_indicator(indicator,
							current, c);
						if (rc) {
							delete_list(list);
							return rc;
						}
					}
				}

				if (!(rc = get_sensor(indicator, current,
						&state))) {
					if (dvalue)
						printf("%s\t[%s]\n", lvalue,
							state?"on":"off");
					else
						printf("%s\n",
							state?"on":"off");
				}
				delete_list(list);
				return rc;
			}

			current = current->next;
		}

		strcpy(temp, lvalue);

		if(trunc) {
			if(truncate_loc_code(lvalue)) {
				truncated = 1;
				goto retry;
			}
		}

		printf("There is no indicator at location code %s\n",
			temp);
		delete_list(list);
		return 2;
	}

	if (othervalue) {
		/* Turn all indicators either on or off */
		if (!strcmp(othervalue, "all-on"))
			c = 1;
		else if (!strcmp(othervalue, "all-off"))
			c = 0;
		else {
			fprintf(stderr, "oops! neither all-on nor all-off"
				"was specified!\n");
			delete_list(list);
			return 1;
		}

		current = list;
		while (current) {
			if (!get_sensor(indicator, current, &state)) {
				if (state == c)
					printf("%s\t[%s]\n", current->code,
						c?"on":"off");
				else {
					set_indicator(indicator, current, c);
					if (!get_sensor(indicator, current,
							&state)) {
						printf("%s\t[%s]\n",
							current->code,
							state?"on":"off");
					}
				}
				fflush(stdout);
			}
			current	= current->next;
		}
	}

	delete_list(list);
	return 0;
}
